#!/usr/bin/env python3

'''
`@beartype` decorator, implementing a rudimentary subset of PEP 484-style type
checking based on Python 3.x function annotations.

See Also
----------
https://stackoverflow.com/a/37961120/2809027
    Stackoverflow answer introducing the `@beartype` decorator.

Source:
https://gist.github.com/terrdavis/1b23b7ff8023f55f627199b09cfa6b24

Credits:
@leycec (author of original)
@terrdavis (fixes for Python 3.7)
'''

# ....................{ MAIN                               }....................
# If the active Python interpreter is *NOT* optimized (e.g., option "-O" was
# *NOT* passed to this interpreter), enable type checking.
if __debug__:
    import inspect
    from functools import wraps
    from inspect import Parameter, Signature

    def beartype(func: callable) -> callable:
        '''
        Decorate the passed **callable** (e.g., function, method) to validate
        both all annotated parameters passed to this callable _and_ the
        annotated value returned by this callable if any.

        This decorator performs rudimentary type checking based on Python 3.x
        function annotations, as officially documented by PEP 484 ("Type
        Hints"). While PEP 484 supports arbitrarily complex type composition,
        this decorator requires _all_ parameter and return value annotations to
        be either:

        * Classes (e.g., `int`, `OrderedDict`).
        * Tuples of classes (e.g., `(int, OrderedDict)`).

        If optimizations are enabled by the active Python interpreter (e.g., due
        to option `-O` passed to this interpreter), this decorator is a noop.

        Raises
        ----------
        NameError
            If any parameter has the reserved name `__beartype_func`.
        TypeError
            If either:
            * Any parameter or return value annotation is neither:
              * A type.
              * A tuple of types.
            * The kind of any parameter is unrecognized. This should _never_
              happen, assuming no significant changes to Python semantics.
        '''

        # Raw string of Python statements comprising the body of this wrapper,
        # including (in order):
        #
        # * A "@wraps" decorator propagating the name, docstring, and other
        #   identifying metadata of the original function to this wrapper.
        # * A private "__beartype_func" parameter initialized to this function.
        #   In theory, the "func" parameter passed to this decorator should be
        #   accessible as a closure-style local in this wrapper. For unknown
        #   reasons (presumably, a subtle bug in the exec() builtin), this is
        #   not the case. Instead, a closure-style local must be simulated by
        #   passing the "func" parameter to this function at function
        #   definition time as the default value of an arbitrary parameter. To
        #   ensure this default is *NOT* overwritten by a function accepting a
        #   parameter of the same name, this edge case is tested for below.
        # * Assert statements type checking parameters passed to this callable.
        # * A call to this callable.
        # * An assert statement type checking the value returned by this
        #   callable.
        #
        # While there exist numerous alternatives (e.g., appending to a list or
        # bytearray before joining the elements of that iterable into a string),
        # these alternatives are either slower (as in the case of a list, due to
        # the high up-front cost of list construction) or substantially more
        # cumbersome (as in the case of a bytearray). Since string concatenation
        # is heavily optimized by the official CPython interpreter, the simplest
        # approach is (curiously) the most ideal.
        func_body = '''
@wraps(__beartype_func)
def func_beartyped(*args, __beartype_func=__beartype_func, **kwargs):
    missing = object()
'''

        # "inspect.Signature" instance encapsulating this callable's signature.
        func_sig = inspect.signature(func)

        # Human-readable name of this function for use in exceptions.
        func_name = func.__name__ + '()'

        # For the name of each parameter passed to this callable and the
        # "inspect.Parameter" instance encapsulating this parameter (in the
        # passed order)...
        for func_arg_index, func_arg in enumerate(func_sig.parameters.values()):
            # If this callable redefines a parameter initialized to a default
            # value by this wrapper, raise an exception. Permitting this
            # unlikely edge case would permit unsuspecting users to
            # "accidentally" override these defaults.
            if func_arg.name == '__beartype_func':
                raise NameError(
                    'Parameter {} reserved for use by @beartype.'.format(
                        func_arg.name))

            # If this parameter is both annotated and non-ignorable for purposes
            # of type checking, type check this parameter.
            if (func_arg.annotation is not Parameter.empty and
                func_arg.kind not in _PARAMETER_KIND_IGNORED):
                # Validate this annotation.
                _check_type_annotation(
                    annotation=func_arg.annotation,
                    label='{} parameter {} type'.format(
                        func_name, func_arg.name))

                # String evaluating to this parameter's annotated type.
                func_arg_type_expr = (
                    "getattr(__beartype_func.__annotations__[{0!r}], '__origin__', None) "
                    "or __beartype_func.__annotations__[{0!r}]".format(func_arg.name))

                # String evaluating to this parameter's current value when
                # passed as a keyword.
                func_arg_value_key_expr = 'kwargs.get({!r}, missing)'.format(func_arg.name)

                # If this parameter is keyword-only, type check this parameter
                # only by lookup in the variadic "**kwargs" dictionary.
                if func_arg.kind is Parameter.KEYWORD_ONLY:
                    func_body += '''
    {arg_name}_type = {arg_type_expr}
    {arg_name}_value = {arg_value_key_expr}
    if {arg_name}_value is not missing and not isinstance({arg_name}_value, {arg_name}_type):
        raise TypeError(
            '{func_name} keyword-only parameter {arg_name}={{}} not a {{!r}}'.format(
                {arg_name}_value, {arg_name}_type))
'''.format(
                        func_name=func_name,
                        arg_name=func_arg.name,
                        arg_type_expr=func_arg_type_expr,
                        arg_value_key_expr=func_arg_value_key_expr,
                    )
                # Else, this parameter may be passed either positionally or as
                # a keyword. Type check this parameter both by lookup in the
                # variadic "**kwargs" dictionary *AND* by index into the
                # variadic "*args" tuple.
                else:
                    # String evaluating to this parameter's current value when
                    # passed positionally.
                    func_arg_value_pos_expr = 'args[{!r}]'.format(
                        func_arg_index)

                    func_body += '''
    type_{arg_index} = {arg_type_expr}
    value_{arg_index} = {arg_value_pos_expr} if {arg_index} < len(args) else {arg_value_key_expr}
    if value_{arg_index} is not missing and not isinstance(value_{arg_index}, type_{arg_index}):
            raise TypeError('{func_name} parameter {arg_name}={{}} not of {{!r}}'.format(
                value_{arg_index}, type_{arg_index}))
'''.format(
                    func_name=func_name,
                    arg_name=func_arg.name,
                    arg_index=func_arg_index,
                    arg_type_expr=func_arg_type_expr,
                    arg_value_key_expr=func_arg_value_key_expr,
                    arg_value_pos_expr=func_arg_value_pos_expr,
                )

        # If this callable's return value is both annotated and non-ignorable
        # for purposes of type checking, type check this value.
        if func_sig.return_annotation not in _RETURN_ANNOTATION_IGNORED:
            # Validate this annotation.
            _check_type_annotation(
                annotation=func_sig.return_annotation,
                label='{} return type'.format(func_name))

            # Strings evaluating to this parameter's annotated type and
            # currently passed value, as above.
            func_return_type_expr = (
                "getattr(__beartype_func.__annotations__['return'], '__origin__', None) "
                "or __beartype_func.__annotations__['return']")

            # Call this callable, type check the returned value, and return this
            # value from this wrapper.
            func_body += '''
    return_value = __beartype_func(*args, **kwargs)
    return_type = {return_type}
    if not isinstance(return_value, return_type):
        raise TypeError('{func_name} return value {{}} not of {{!r}}'.format(return_value, return_type))
    return return_value
'''.format(func_name=func_name, return_type=func_return_type_expr)
        # Else, call this callable and return this value from this wrapper.
        else:
            func_body += '''
    return __beartype_func(*args, **kwargs)
'''

        # Dictionary mapping from local attribute name to value. For efficiency,
        # only those local attributes explicitly required in the body of this
        # wrapper are copied from the current namespace. (See below.)
        local_attrs = {'__beartype_func': func}

        # Dynamically define this wrapper as a closure of this decorator. For
        # obscure and presumably uninteresting reasons, Python fails to locally
        # declare this closure when the locals() dictionary is passed; to
        # capture this closure, a local dictionary must be passed instead.
        exec(func_body, globals(), local_attrs)

        # Return this wrapper.
        return local_attrs['func_beartyped']

    _PARAMETER_KIND_IGNORED = {
        Parameter.POSITIONAL_ONLY, Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD,
    }
    '''
    Set of all `inspect.Parameter.kind` constants to be ignored during
    annotation- based type checking in the `@beartype` decorator.

    This includes:

    * Constants specific to variadic parameters (e.g., `*args`, `**kwargs`).
      Variadic parameters cannot be annotated and hence cannot be type checked.
    * Constants specific to positional-only parameters, which apply to non-pure-
      Python callables (e.g., defined by C extensions). The `@beartype`
      decorator applies _only_ to pure-Python callables, which provide no
      syntactic means of specifying positional-only parameters.
    '''

    _RETURN_ANNOTATION_IGNORED = {Signature.empty, None}
    '''
    Set of all annotations for return values to be ignored during annotation-
    based type checking in the `@beartype` decorator.

    This includes:

    * `Signature.empty`, signifying a callable whose return value is _not_
      annotated.
    * `None`, signifying a callable returning no value. By convention, callables
      returning no value are typically annotated to return `None`. Technically,
      callables whose return values are annotated as `None` _could_ be
      explicitly checked to return `None` rather than a none-`None` value. Since
      return values are safely ignorable by callers, however, there appears to
      be little real-world utility in enforcing this constraint.
    '''

    def _check_type_annotation(annotation: object, label: str) -> None:
        '''
        Validate the passed annotation to be a valid type supported by the
        `@beartype` decorator.

        Parameters
        ----------
        annotation : object
            Annotation to be validated.
        label : str
            Human-readable label describing this annotation, interpolated into
            exceptions raised by this function.

        Raises
        ----------
        TypeError
            If this annotation is neither a new-style class nor a tuple of
            new-style classes.
        '''

        # If this annotation is a tuple, raise an exception if any member of
        # this tuple is not a new-style class. Note that the "__name__"
        # attribute tested below is not defined by old-style classes and hence
        # serves as a helpful means of identifying new-style classes.
        if isinstance(annotation, tuple):
            for member in annotation:
                if not (
                    hasattr(member, '__origin__') or
                    isinstance(member, type) and hasattr(member, '__name__')):
                    raise TypeError(
                        '{} tuple member {} not a new-style class'.format(
                            label, member))
        # Else if this annotation is not a new-style class, raise an exception.
        elif not (
            hasattr(annotation, '__origin__') or
            isinstance(annotation, type) and hasattr(annotation, '__name__')):
            raise TypeError(
                '{} {} neither a new-style class nor '
                'tuple of such classes'.format(label, annotation))

# Else, the active Python interpreter is optimized. In this case, disable type
# checking by reducing this decorator to the identity decorator.
else:
    def beartype(func: callable) -> callable:
        return func


"""A contiguous frame cache data structure, designed to support efficient data loading
from video for the setting in which:
    * frame accesses are sequential (i.e. not random access)
    * the same frames can be read multiple times

Motivation: Suppose the frames of a video are labelled
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]
The frame cache aims to store frames in memory so that it is fast to produce batches that
repeat frames e.g.
    [0, 1, 2, 3, 4],
       [1, 2, 3, 4, 5],
          [2, 3, 4, 5, 6],

It does this by allocating a single cache of shared storage, and loading frames into it.
Repeated local frame access are then fast, provided they fall within the cache. It is also
possible to access strided batches e.g:
    [0, 1, 2, 3, 4],
       [2, 3, 4, 5, 6],
          [4, 5, 6, 7, 8],
However, note that the underlying frame remains contiguous (i.e. in this case, it will
still store [0, 1, 2, 3, 4, 5, 6,.....])
"""
import argparse
from typing import Tuple
from pathlib import Path

import numpy as np
from beartype import beartype
from beartype.cave import NoneTypeOr


class DummyVideoReader:
    """A dummy video reader that mimics the OpenCV VideoCapture interface, but yields
    integers rather than frames. Useful for sanity checking frame indices.
    """

    def __init__(self, total_video_frames: int):
        """Create an instance of DummyVideoReader

        Args:
            total_video_frames: the total number of frames to be used in the dummy video.
        """
        self.ptr = 0
        self.frames = list(range(total_video_frames))

    def read(self) -> Tuple:
        """Mimic the OpenCV VideoCapture read() function.

        Returns:
            A tuple comprising a boolean indicating whether the dummy frame read was
            successful, and integer corresponding to the total number of times the read()
            function has been called. If read() is called more times than the total number
            of dummy video frames, the return code is False and the return value is None
        """
        if self.ptr < len(self.frames):
            val = self.frames[self.ptr]
            self.ptr += 1
        else:
            val = None
        return_code = val is not None
        return return_code, val


class FrameOutOfBoundsError(ValueError):
    """A cache exception that will be raised whenever the user requests frame sequences
    that are not legal for the underlying video."""


class CacheCapacityError(ValueError):
    """A cache exception that will be raised whenever the user requests frame sequences
    that cannot be retrieved due to the limited capacity of the cache."""


class InvalidFrameSequence(ValueError):
    """A cache exception that will be raised whenever the user requests an invalid
    frame sequence."""


class ContigFrameCache:
    """A cache to support efficient sequential access to video frames.
    """

    @beartype
    def __init__(
            self,
            num_cache_frames: int,
            frame_height: int,
            frame_width: int,
            cache_dtype: type,
            video_path: Path,
            total_video_frames: int,
            frame_channels: int = 3,
            channel_order: NoneTypeOr[str] = "rgb",
            pad_last: str = "no_pad",
            verbose: bool = False,
            backend: str = "cv2",
    ):
        """Create an instance of ContigFrameCache

        Args:
            num_cache_frames: the number of frames to be stored in the cache at a time
            frame_height: the height (in pixels) of frames to be stored in the cache
            frame_width: the width (in pixels) of frames to be stored in the cache
            cache_dtype: the data type to be used by the frame cache
            video_path: the location of the video to be read
            total_video_frames: the length of the video to be read (in frames)
            frame_channels: the number of channels for each frame (e.g. 3 for RGB)
            channel_order: if three channels are requested, store frames in this order. Can
                be one of None, "rgb" or "bgr" (the latter being the opencv default).
            pad_last: how to handle frame requests beyond the total number of frames in the
                video. Options are `no_pad`, in which case an exception will be raised, or
                `copy_last` in which case the final frame of the video will be duplicated to
                fill the requested frame range.
            verbose: whether to print out detailed information during operation.
            backend: the backend to use as a video reader. Can be either `dummy` or `cv2`
        """
        self.storage = np.zeros((num_cache_frames, frame_height, frame_width,
                                frame_channels), dtype=cache_dtype)

        assert video_path.exists(), f"Could not find video at {video_path}"
        if backend == "dummy":
            assert frame_channels == 1, "dummy backend should use a single channel"
            self.video_cap = DummyVideoReader(total_video_frames)
        elif backend == "cv2":
            try:
                import cv2  # pylint: disable = import-outside-toplevel
            except ImportError:
                raise ImportError("To use cv2 backend, opencv must be installed") from None
            self.video_cap = cv2.VideoCapture(str(video_path))
        else:
            raise ValueError(f"Unknown backend: {backend}")
        if pad_last not in {"no_pad", "zero_pad", "copy_last"}:
            raise ValueError(f"Unknown pad_last strategy: {pad_last}")
        if channel_order not in {None, "rgb", "bgr"}:
            raise ValueError(f"Unknown order: {channel_order}")
        elif channel_order in {"rgb", "bgr"} and frame_channels != 3:
            raise ValueError(f"Expected 3 channels {frame_channels} for {channel_order}")

        self.num_cache_frames = num_cache_frames
        self.channel_order = channel_order
        self.frame_height = frame_height
        self.total_video_frames = total_video_frames
        self.frame_width = frame_width
        self.frame_channels = frame_channels
        self.pad_last = pad_last
        self.verbose = verbose
        self.head_ptr = None

    @beartype
    def query(self, start_frame: int, end_frame: int) -> bool:
        """Determine whether the sequence of frames [start_frame, end_frame) is contained
        within the frame cache.

        Args:
            start_frame: starting frame in the sequence
            end_frame: end frame in the sequence

        Returns:
            True if the sequence falls within the cache, False otherwise.
        """
        if end_frame - start_frame > self.num_cache_frames:
            raise CacheCapacityError(f"Requested a sequence of {end_frame - start_frame} "
                                     "frames (larger than cache size of "
                                     f"{self.num_cache_frames} frames)")

        if end_frame - start_frame <= 0:
            raise InvalidFrameSequence("Number of frames must be non-negative, "
                                       f"requested {end_frame - start_frame}")

        if start_frame < 0:
            raise InvalidFrameSequence("Starting frame of segment must be non-negative, "
                                       f"requested {start_frame}")

        if self.pad_last == "no_pad" and end_frame > self.total_video_frames:
            raise FrameOutOfBoundsError("When no_pad strategy is used cannot request end "
                                        f"frame {end_frame} beyond the total video length "
                                        f" ({self.total_video_frames} frames)")
        if (
            self.pad_last in {"zero_pad", "copy_last"}
                and start_frame >= self.total_video_frames
        ):
            raise FrameOutOfBoundsError(f"When {self.pad_last} strategy is used cannot "
                                        f"request start frame {start_frame} beyond the "
                                        f" total video length ({self.total_video_frames} "
                                        "frames)")

        if self.head_ptr is None:
            return False

        last_cache_frame = self.head_ptr + self.num_cache_frames
        if self.pad_last == "no_pad":
            hit = start_frame >= self.head_ptr and end_frame <= last_cache_frame
        elif self.pad_last in {"zero_pad", "copy_last"}:
            hit = self.head_ptr <= start_frame <= last_cache_frame
        return hit

    @beartype
    def fetch(self, start_frame: int, end_frame: int, copy: bool = True) -> np.ndarray:
        """Fetch a sequence of frames, updating the cache if necessary.

        Args:
            start_frame: starting frame in the requested sequence
            end_frame: end frame in the requested sequence
            copy: whether to return a copy of cached frames, or a direct reference to the
                underlying data.

        Returns:
            The corresponding frame sequence.
        """
        while not self.query(start_frame, end_frame):
            self._update(start_frame)

        start_idx = start_frame - self.head_ptr
        end_idx = end_frame - self.head_ptr
        frames = self.storage[start_idx:end_idx]
        if copy:
            frames = frames.copy()
        return frames

    @beartype
    def _update(self, start_frame: int):
        """Update the frame storage cache so that it's contents begin at the given
        start_frame.

        Args:
            start_frame: the frame at which the updated cache will begin
        """
        if self.verbose:
            print(f"Re-aligning cache to begin at frame {start_frame}")
        if self.head_ptr is None:
            # fill the empty cache
            start_idx = 0
            num_fill_frames = self.num_cache_frames
            remaining_frames_in_video = self.total_video_frames

        elif start_frame <= self.head_ptr + self.num_cache_frames:
            # shift cache frames so that storage begins with the requested frame
            offset = start_frame - self.head_ptr
            keep = self.head_ptr + self.num_cache_frames - start_frame
            self.storage[0:keep] = self.storage[offset:]
            start_idx = keep
            num_fill_frames = self.num_cache_frames - keep
            remaining_frames_in_video = self.total_video_frames - (start_frame + keep)

        else:
            # we need to skip frames in order to refill the cache
            num_skip = start_frame - (self.head_ptr + self.num_cache_frames)
            self.skip_frames(num_skip)

            # refill the cache
            start_idx = 0
            num_fill_frames = self.num_cache_frames
            remaining_frames_in_video = self.total_video_frames - start_frame

        if self.pad_last == "no_pad":
            num_fill_frames = min(num_fill_frames, remaining_frames_in_video)

        self._fill_cache(start_idx=start_idx, num_frames=num_fill_frames)
        self.head_ptr = start_frame

    @beartype
    def skip_frames(self, num_frames: int):
        """Skip forwards by the given number of frames in the video.

        Args:
            The number of frames to skip
        """
        if self.verbose:
            print(f"Skipping {num_frames} frames")
        for _ in range(num_frames):
            self.video_cap.read()

    @beartype
    def compute_num_sequences(
            self,
            sequence_length: int,
            sequence_stride: int,
    ) -> int:
        """A utility function to compute the total number of sequences that will be produced
        by slicing the frames of the video associated with the current cache into sequences.

        Args:
            sequence_length: the length of each sequence in frames.
            sequence_stride: the stride (number of frames between the start of each
                sequence).

        Returns:
            the total number of sequences that will be produced by the video.
        """
        if sequence_length <= 0:
            raise InvalidFrameSequence("Cannot compute number of sequences for non-positive"
                                       " sequence length")
        if sequence_stride <= 0:
            raise ValueError("Cannot compute number of sequences for non-positive"
                             " sequence stride")

        # handle the edge case in which the video is empty
        if self.total_video_frames == 0:
            return 0

        if self.pad_last == "no_pad":
            # handle the edge case in which the requested sequence is longer than the video
            # itself
            if sequence_length > self.total_video_frames:
                num_seqs = 0
            else:
                # handle more typical cases
                num_seqs = 1 + (self.total_video_frames - sequence_length) / sequence_stride
                num_seqs = int(np.floor(num_seqs))

        elif self.pad_last in {"zero_pad", "copy_last"}:
            num_seqs = int(np.ceil(self.total_video_frames / sequence_stride))
        return num_seqs

    @beartype
    def _fill_cache(self, start_idx: int, num_frames: int):
        """Fill the contents of the cache storage with the next sequence of frames from
        the video.

        Args:
            start_idx: the position in cache storage at which to begin writing new frames.
            num_frames: the number of frames to write into storage.
        """
        # We keep track of the previous frame loaded from the video to ensure that we can
        # implement 'copy_last' frame padding if requested
        prev_im = None

        for frame_idx in range(num_frames):

            _, im = self.video_cap.read()

            if im is None:
                if self.pad_last == "no_pad":
                    raise FrameOutOfBoundsError(
                        f"Attempted to read frame {frame_idx + start_idx + self.head_ptr}, "
                        " beyond the end of the video"
                    )
                elif self.pad_last == "zero_pad":
                    im = prev_im * 0
                elif self.pad_last == "copy_last":
                    im = prev_im

            if self.channel_order == "rgb":
                im = im[:, :, ::-1]
            elif self.channel_order == "bgr":
                pass

            self.storage[start_idx + frame_idx] = im

            if im is not None:
                prev_im = im


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--video_path", type=Path, default="data/dali_mwp/sintel.mp4")
    parser.add_argument("--test_vid_height", type=int, default=256)
    parser.add_argument("--test_vid_width", type=int, default=256)
    parser.add_argument("--num_cache_frames", type=int, default=20)
    parser.add_argument("--total_video_frames", type=int, default=1440)
    parser.add_argument("--pad_last", type=str, default="copy_last")
    parser.add_argument("--batch_size", type=int, default=8)
    parser.add_argument("--backend", type=str, choices=["cv2", "dummy"], default="cv2")
    parser.add_argument("--verbose", action="store_true")
    args = parser.parse_args()

    cache = ContigFrameCache(
        num_cache_frames=args.num_cache_frames,
        frame_height=args.test_vid_height,
        frame_width=args.test_vid_width,
        video_path=args.video_path,
        total_video_frames=args.total_video_frames,
        cache_dtype=np.uint8,
        verbose=args.verbose,
        pad_last=args.pad_last,
        backend=args.backend,
    )

    for batch_start in range(0, args.total_video_frames + 1, 1):
        batch_end = batch_start + args.batch_size
        frames_ = cache.fetch(start_frame=batch_start, end_frame=batch_end)
        print(f"Batch offset: {batch_start}, frames: {frames_.shape}")
        cache_content = [x[0][0][0] for x in cache.storage]
        batch_content = [x[0][0][0] for x in frames_]
        print("Batch content: " + ",".join(str(x) for x in batch_content))
        print("Cache content: " + ",".join(str(x) for x in cache_content))



import sys
import json
import time
import pickle
import socket
import numbers
import functools
import logging
import unicodedata
import subprocess
from typing import Dict, List, Union
from pathlib import Path
import yaml

import numpy as np
import hickle
import scipy.io as spio
import msgpack_numpy as msgpack_np
from beartype import beartype
from mergedeep import Strategy, merge
from typeguard import typechecked
from beartype.cave import AnyType, NoneTypeOr
import zsvision.zs_data_structures


@functools.lru_cache(maxsize=64, typed=False)
@beartype
def memcache(path: Union[Path, str], verbose: bool = True):
    path = Path(path)
    suffix = path.suffix
    if verbose:
        print(f"loading data from {path} ({socket.gethostname()})", end=" ", flush=True)
    tic = time.time()
    if suffix in {".pkl", ".pickle", ".pckl", ".pk"}:
        res = pickle_loader(pkl_path=path, verbose=verbose)
    elif suffix in {".hkl", ".hickle"}:
        res = hickle.load(path)
    elif suffix == ".npy":
        res = np_loader(path, verbose=verbose)
    elif suffix == ".mp":
        res = msgpack_loader(path, verbose=verbose)
    elif suffix == ".json":
        with open(path, "r") as f:
            res = json.load(f)
    elif suffix in {".yaml", ".yml"}:
        with open(path, "r") as f:
            res = yaml.safe_load(f)
    elif suffix == ".mat":
        res = loadmat(path)
    else:
        raise ValueError(f"unknown suffix: {suffix} for path {path}")
    if verbose:
        print(f"[Total: {time.time() - tic:.1f}s]")
    return res


@beartype
def support_old_pickles(buffer: bytes) -> object:
    try:
        data = pickle.loads(buffer, encoding="latin1")
    except ModuleNotFoundError as exception:
        if "datastructures" in str(exception.msg):
            sys.modules['datastructures'] = zsvision.zs_data_structures
            data = pickle.loads(buffer, encoding="latin1")
    return data


@beartype
def pickle_loader(
        pkl_path: Path,
        verbose: bool,
        backwards_compatible: bool = True,
) -> object:
    """Deserialise object from pickle.

    Args:
        pkl_path: the location of the path where the pickle path is stored
        backwards_compatible: if true, support old pickle formats used with the.
            ExpertStore format

    Return:
        The deserialised object.
    """
    tic = time.time()
    with open(pkl_path, "rb") as f:
        buffer = f.read()
        if verbose:
            print(f"[I/O: {time.time() - tic:.1f}s]", end=" ")
        tic = time.time()
        if backwards_compatible:
            data = support_old_pickles(buffer)
        else:
            data = pickle.loads(buffer, encoding="latin1")
        if verbose:
            print(f"[deserialisation: {time.time() - tic:.1f}s]", end=" ")
    return data


@beartype
def msgpack_loader(mp_path: Path, verbose: bool):
    """Msgpack provides a faster serialisation routine than pickle, so is preferable
    for loading and deserialising large feature sets from disk."""
    tic = time.time()
    with open(mp_path, "rb") as f:
        buffer = f.read()
        if verbose:
            print(f"[I/O: {time.time() - tic:.1f}s]", end=" ")
        tic = time.time()
        data = msgpack_np.unpackb(buffer, raw=False)
        if verbose:
            print(f"[deserialisation: {time.time() - tic:.1f}s]", end=" ")
    return data


@beartype
def np_loader(np_path: Path, verbose: bool, l2norm: bool = False):
    with open(np_path, "rb") as f:
        data = np.load(f, encoding="latin1", allow_pickle=True)
    if isinstance(data, np.ndarray) and data.size == 1:
        data = data[()]  # handle numpy dict storage convnetion
    if l2norm:
        if verbose:
            print("L2 normalizing features")
        if isinstance(data, dict):
            for key in data:
                feats_ = data[key]
                feats_ = feats_ / max(np.linalg.norm(feats_), 1E-6)
                data[key] = feats_
        elif data.ndim == 2:
            data_norm = np.linalg.norm(data, axis=1)
            data = data / np.maximum(data_norm.reshape(-1, 1), 1E-6)
        else:
            raise ValueError("unexpected data format {}".format(type(data)))
    return data


@beartype
def set_nested_key_val(key: str, val: AnyType, target: dict):
    """Use a prefix key (e.g. key1.key2.key3) to set a value in a nested dict"""

    # escape periods in keys
    key = key.replace("_.", "&&")
    subkeys = key.split(".")
    subkeys = [x.replace("&&", ".") for x in subkeys]

    nested = target
    print("subkeys", subkeys)
    for subkey in subkeys[:-1]:
        try:
            nested = nested.__getitem__(subkey)
        except Exception as exception:
            print(subkey)
            raise exception
    orig = nested[subkeys[-1]]
    if orig == "":
        if val == "":
            val = 0
        else:
            val = str(val)
    elif isinstance(orig, bool):
        if val.lower() in {"0", "False"}:
            val = False
        else:
            val = bool(val)
    elif isinstance(orig, list):
        if isinstance(val, str) and "," in val:
            val = val.split(",")
            # we use the convention that a trailing comma indicates a single item list
            if len(val) == 2 and val[1] == "":
                val.pop()
            if val and not orig:
                raise ValueError(f"Could not infer correct type from empty original list")
            else:
                val = [type(orig[0])(x) for x in val]
        assert isinstance(val, list), "Failed to pass a list where expected"
    elif isinstance(orig, int):
        val = int(val)
    elif isinstance(orig, float):
        val = float(val)
    elif isinstance(orig, str):
        val = str(val)
    else:
        raise ValueError(f"unrecognised type: {type(val)}")
    nested[subkeys[-1]] = val


@beartype
def loadmat(src_path: Path) -> dict:
    """This function should be called instead of direct spio.loadmat as it addresses the
    problem of not properly recovering python dictionaries from mat files. It calls the
    function check keys to cure all entries which are still mat-objects.

    The function is heavily based on this reference:
    https://stackoverflow.com/a/29126361

    Args:
        src_path: the location of the .mat file to load

    Returns:
        a parsed .mat file in the form of a python dictionary.
    """
    def _check_keys(d):
        """Checks if entries in dictionary are mat-objects. If yes
        todict is called to change them to nested dictionaries
        """
        for key in d:
            if isinstance(d[key], spio.matlab.mio5_params.mat_struct):
                d[key] = _todict(d[key])
            elif isinstance(d[key], np.ndarray):
                d[key] = _tolist(d[key])
            else:
                pass
        return d

    def _todict(matobj):
        """A recursive function which constructs from matobjects nested dictionaries
        """
        d = {}
        for strg in matobj._fieldnames:
            elem = matobj.__dict__[strg]
            if isinstance(elem, spio.matlab.mio5_params.mat_struct):
                d[strg] = _todict(elem)
            elif isinstance(elem, np.ndarray):
                d[strg] = _tolist(elem)
            else:
                d[strg] = elem
        return d

    def _tolist(ndarray):
        """A recursive function which constructs lists from cellarrays
        (which are loaded as numpy ndarrays), recursing into the elements
        if they contain matobjects or are non-numeric.
        """
        if np.issubdtype(ndarray.dtype, np.number):
            return ndarray
        elem_list = []
        for sub_elem in ndarray:
            if isinstance(sub_elem, spio.matlab.mio5_params.mat_struct):
                elem_list.append(_todict(sub_elem))
            elif isinstance(sub_elem, np.ndarray):
                elem_list.append(_tolist(sub_elem))
            else:
                elem_list.append(sub_elem)
        return elem_list

    data = spio.loadmat(src_path, struct_as_record=False, squeeze_me=True)
    return _check_keys(data)


@functools.lru_cache(maxsize=64, typed=False)
def concat_features(feat_paths, axis):
    aggregates = [memcache(x) for x in feat_paths]
    tic = time.time()
    msg = "expected to concatenate datastructures of a single type"
    assert len(set(type(x) for x in aggregates)) == 1, msg
    if isinstance(aggregates[0], dict):
        keys = aggregates[0]  # for now, we assume that all aggregates share keys
        merged = {}
        for key in keys:
            merged[key] = np.concatenate([x[key] for x in aggregates], axis=axis)
    elif isinstance(aggregates[0], zsvision.zs_data_structures.ExpertStore):
        dims, stores = [], []
        keys = aggregates[0].keys
        for x in aggregates:
            dims.append(x.dim)
            stores.append(x.store)
            try:
                assert x.keys == keys, "all aggregates must share identical keys"
            except Exception as E:
                print(E)
                import ipdb; ipdb.set_trace()
        msg = "expected to concatenate ExpertStores with a common dimension"
        assert len(set(dims)) == 1, msg
        dim = dims[0]
        merged = zsvision.zs_data_structures.ExpertStore(keys, dim=dim)
        merged.store = np.concatenate(stores, axis=axis)
    else:
        raise ValueError(f"Unknown datastructure: {type(aggregates[0])}")
    # Force memory clearance
    for aggregate in aggregates:
        del aggregate
    print("done in {:.3f}s".format(time.time() - tic))
    return merged


class BlockTimer:
    """A minimal inline codeblock timer"""
    @beartype
    def __init__(
            self,
            msg: str,
            mute: bool = False,
            precise: bool = False,
            logger: NoneTypeOr[logging.Logger] = False,
    ):
        self.msg = msg
        self.mute = mute
        self.precise = precise
        self.logger = logger
        self.start = None

    def __enter__(self):
        self.start = time.time()
        msg = f"{self.msg}..."
        if self.logger:
            self.logger.info(msg)
        else:
            print(msg, end="", flush=True)
        return self

    def __exit__(self, *args):
        if self.precise:
            total = f"{time.time() - self.start:.3f}s"
        else:
            total = time.strftime('%Hh%Mm%Ss', time.gmtime(time.time() - self.start))
        if not self.mute:
            msg = f" took {total}"
            if self.logger:
                self.logger.info(msg)
            else:
                print(msg)


@beartype
def find_ancestors(cfg_fname: (Path, str)) -> list:
    """Search the hierarchy specified by the `inherit_from` attribute of a json config
    via post-order traversal.

    Args:
        cfg_fname: the location of the json config file

    Returns:
        a list of loaded configs in the order specified by the inheritance.
    """
    # Cannot use memcache here without risk of recursion
    if Path(cfg_fname).suffix == ".json":
        with open(cfg_fname, "r") as f:
            config = json.load(f)
    elif Path(cfg_fname).suffix in {".yaml", ".yml"}:
        with open(cfg_fname, "r") as f:
            config = yaml.safe_load(f)
    else:
        raise ValueError(f"Unknown config path type: {cfg_fname}")
    ancestors = []
    if "inherit_from" in config:
        immediate_ancestors = config["inherit_from"].split(",")
        for immediate_ancestor in immediate_ancestors:
            ancestors.extend(find_ancestors(Path(immediate_ancestor)))
    ancestors.append(config)
    return ancestors


@beartype
def load_json_or_yaml_config(cfg_fname: (Path, str)) -> dict:
    """Load a configuration file into memory.

    Args:
        cfg_fname: the location of the config file (yaml or json)

    Returns:
        the loaded configuration
    """
    ancestors = find_ancestors(cfg_fname)
    config = ancestors.pop()
    ancestors = reversed(ancestors)
    for ancestor in ancestors:
        merge(ancestor, config, strategy=Strategy.REPLACE)
        config = ancestor
    return config


@beartype
def load_json_config(cfg_fname: (Path, str)) -> dict:
    """Load a json configuration file into memory.

    Args:
        cfg_fname: the location of the json config file

    Returns:
        the loaded configuration

    NOTES: A json file may include an `inherit_from`: "<path>" key, value pair which
    points to a list of templates from which to inherit default values.  Inheritance
    specifiers are traversed in increasing order of importance, from left to right.
    E.g. given
        "inherit_from": "path-to-A,path-to-B",
    the values of B will override the values of A.
    """
    return load_json_or_yaml_config(cfg_fname)


@beartype
def load_yaml_config(cfg_fname: (Path, str)) -> dict:
    """Load a yaml configuration file into memory.

    Args:
        cfg_fname: the location of the yaml config file

    Returns:
        the loaded configuration

    NOTES: A yaml file may include an `inherit_from`: "<path>" key, value pair which
    points to a list of templates from which to inherit default values.  Inheritance
    specifiers are traversed in increasing order of importance, from left to right.
    E.g. given
        "inherit_from": "path-to-A,path-to-B",
    the values of B will override the values of A.
    """
    return load_json_or_yaml_config(cfg_fname)


@beartype
def seconds_to_timestr(secs: numbers.Number) -> str:
    """Convert a total number of seconds into a formatted time string.

    Arguments:
        secs: the total number of seconds

    Returns:
        a formatted time (HH:MM:SS.mmm)

    NOTE: Probably this function is not needed. But I refuse to spend more of my life
    looking at datetime/time/strftime combinations.
    """
    assert secs >= 0, "Expected a non-negative number of seconds"
    mins, secs = divmod(secs, 60)
    hours, mins = divmod(mins, 60)
    ms = secs - int(secs)
    return f"{int(hours):02d}:{int(mins):02d}:{int(secs):02d}.{int(ms * 1000):03d}"


@typechecked
def list_visible_gpu_types() -> List[str]:
    """Provide a list of the NVIDIA GPUs that are visible on the current machine.

    Returns:
        a list of GPU device types.
    """
    cmd = ["nvidia-smi", "-L"]
    try:
        res = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                             check=True)
        device_strs = res.stdout.decode("utf-8").splitlines()
        devices = [x.split(":")[1].split("(")[0].strip() for x in device_strs]
    except FileNotFoundError:
        devices = []
    return devices


@beartype
def quote_and_escape_ffmpeg_path(path: (str, Path)) -> str:
    """Quote and escape paths for use with ffmpeg/ffprobe.

    Args:
        path: the location of a file to be processed by ffmpeg

    Returns:
        a quoted, dollar-escaped path

    Example usage:
        `os.system("ffprobe {quote_and_escape_ffmpeg_path(path)}")`

    NOTE: This function is useful for processing file paths that may contain:
        1. spaces
        2. dollar characters ($)
        3. percent sign characters (%)
    when invoking ffmpeg or ffprobe from python.
    """
    # Dollar signs need to be escaped when used in paths
    escaped = str(path).replace("$", r"\$").replace("%", r"\%")
    if "'" in escaped:
        quoted = f'"{escaped}"'
    else:
        quoted = f"'{escaped}'"
    return quoted


@beartype
def parse_tree_layout(
        tree_layout_path: Path,
        prefix_token: str = "── ",
) -> set:
    """Given a text dump of the output of the linux `tree` command, this function will
    reconstruct the relative paths of the files in the tree.

    Args:
        tree_layout_path: the location of the text file containing the `tree` output
        prefix_token: the token used by the `tree` command to denote a new file.

    Returns:
        the collection of parsed paths.

    NOTES:
    1. This function assumes that it is parsing the output of the tree command that
    has been run in the directory of the structure it is displaying (i.e. `tree` is run)
    without arguments.
    2. The output of each row in the `tree` command is prefixed by a T-bar or an L-bar
    (see example formats (1) and (2) resp. below).
    3. If the file at `tree_layout_path` contains any rows that are not part of the tree
    output, they are ignored.

    Example:
        Given tree outputs of the forms (1) or (2) shown below:

        (1)
            ├── Conversation
            │   ├── Belfast
            │   │   ├── 11+12

        (2)
            └── Conversation
                └── Belfast
                    └── 11+12

        in both cases, this function will return a set of pathlib paths of the form:

        {
            "."
            "Conversation",
            "Conversation/Belfast",
            "Conversation/Belfast/11+12",
        }

    """
    with open(tree_layout_path, "r") as f:
        rows = f.read().splitlines()

    # filter the input to only contain the file tree structure by searching for the
    # presence of the tree prefix token
    rows = [x for x in rows if prefix_token in x]

    # convert nbsp escape codes into spaces
    rows = [unicodedata.normalize("NFKD", x) for x in rows]

    current_path = Path(".")
    paths = {current_path}
    known_prefix_heads = {"├", "└"}
    for row in rows:
        prefix, name = row.split(prefix_token)
        prefix, prefix_head = prefix[:-1], list(prefix).pop(-1)
        msg = f"Expected prefix head to be in {known_prefix_heads} found {prefix_head}"
        assert prefix_head in known_prefix_heads, msg
        assert len(prefix) % 4 == 0, "Expected prefix string length to be a multiple of 4"
        depth = int(len(prefix) / 4)
        current_path = Path(*current_path.parts[:depth]) / name
        paths.add(current_path)
    return paths


if __name__ == "__main__":
    print(list_visible_gpu_types())
